unit MultiWav;

{ MultiWav (C) by Vladimir Kladov, 2005. Version for KOL (can be used in VCL
  projects though. DirectX required.)

  MultiWavObj: PMultiWav allows to play several wave files simultaneously
  (mixed). Waves can be played from file, resource or memory. It is possible
  to control Pane, Fade in/out, Volume of each playing wave while playing.

  Restrictions: only PCM waves are allowed (no ADPCM or other compresed).
    No restriction at all on count of waves played simultaneously (except
    CPU performance).

  Advices:
  It is better to provide waves with bitrate 44100, 22050 or 11025 (quality
  of resampling can be not good, and speed is the best for three formats listed).
  Also, 8 bit waves can be sound scratch a bit when mixed, the best way to avoid
  this is to convert waves to 16 bit (even mono).

  8 Jun 2005
  Todo: now implemented only for case when 44100 stereo 16 bit format directX
  sound buffer is created OK. For other cases it is required additional
  conversion of the mixed data before out it (yet not implemented).

\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\
mailto:   bonanzas@online.sinor.ru
Web-page: http://bonanzas.rinet.ru
}

interface

{$I KOLDEF.INC}
{$O+}

uses Windows, KOL, DirectSound, MMSystem {$IFDEF ERR}, err {$ENDIF};

var
  BufFrames: Integer = 75; // 3 seconds

type
  TWaveSpeed = ( wsNormal, wsSlow2, wsSlow4, wsFast2, wsFast4 );

  TWaveFileHeader = packed record
    lRIFF: array[ 0..3 ] of Char;
    lSize: Integer;
    lWave: array[ 0..3 ] of Char;
    lFormat: array[ 0..3 ] of Char;
    lFmtLen: Integer;
  end;

  TWaveChunk = packed record
    lType: Integer;
    lLen: Integer;
  end;

  PMultiWav = ^TMultiWav;

  PWave = ^TWave;
  TWave = object( TObj )
  private
    fMultiWav: PMultiWav;
    fSpeed: TWaveSpeed;
    fPaused: Boolean;
    fNextWave: PWave;
    procedure SetPaused(const Value: Boolean);
  protected
    fMem: PStream;
    fFile: THandle;
    fSize: Integer;
    FFadeCount: Integer;
    FFadeDelta: Integer;
    fStartPosition, fStartSize: DWORD;
    fWaveFormat: TWaveFormatEx;
    fWavBuffer: array[ 0..20479 ] of WORD;
    procedure Close;
    function CheckSignature( const Signature: TWaveFileHeader ): Boolean;
    function CheckWaveFormatPCM: Boolean;
    function LoadRsrc( Inst: Integer; RsrcName: PChar ): Boolean;
    function LoadFile( const Filename: String ): Boolean;
    function LoadMemory: Boolean;
    procedure ReadSingleBuffer; // always as 44100 Hz stereo 16 bit
      virtual;
  public
    Pan: Integer;
    Volume: Integer;
    BufferActualSize: Integer;
    Loop: Boolean;
    OnEof: TOnEvent;
    property Paused: Boolean read fPaused write SetPaused;
    property NextWave: PWave read fNextWave write fNextWave;
    destructor Destroy; virtual;
    function OK: Boolean;
    function Eof: Boolean;
    function SetFade( FadeDelta, FadeCount: Integer ): PWave;
    procedure Stop; // do not use Free or Destroy! Use Stop!
    property Speed: TWaveSpeed read fSpeed write fSpeed;
    procedure Restart;
    procedure PlayRange( FromSample, Samples: DWORD );
  public
    property StartPosition: DWORD read fStartPosition;
    property StartSize: DWORD read fStartSize;
    function WaveFormat: TWaveFormatEx;
  end;

  TMultiWav = object( TObj )
  private
    FEnabled: Boolean;
    FPrimaryBuffer: Boolean;
    fExclusiveSnd: Boolean;
    fOnly8BitSound: Boolean;
    fPrty: Integer;
    procedure SetPrty(const Value: Integer);
  protected
    procedure SetEnabled(const Value: Boolean);
  public
    DS: IDirectSound;
    SB: IDirectSoundBuffer;
    BufSize: Integer;
    WFormat: TWaveFormatEx;
    Sampling: Boolean;
    LastValue: Byte;
    Buffer: PByte;
    BufLen: DWORD;
    procedure Init; virtual;
    function CreateSndBuffer( var Buf: IDirectSoundBuffer;
             nChan, nHerz, nBits: Integer; var Format: TWaveFormatEx ): Boolean;
    destructor Destroy; virtual;
    procedure Stop;
    procedure StartSampling;
    procedure StartSample;
    procedure GenerateSample;
    function OK: Boolean;
    property Enabled: Boolean read FEnabled write SetEnabled;
    procedure LogFile1( Value: DWORD );
  protected
    CPosition: DWORD;
    BaseByte: DWORD;
    Stopped: Boolean;
    function FillBuffer_16: Boolean;
  protected
    fDestroying: Boolean;
    fThread: PThread;
    fCrit: TRTLCriticalSection;
    fWaves: PList; // list of PWave
    IgnorePosition, CheckPosition: Integer;
    function ThreadExecute( Sender: PThread ): Integer;
  public
    OnEndFrame: TOnEvent;
    property ExclusiveSound: Boolean read fExclusiveSnd write fExclusiveSnd;
    property Only8BitSound: Boolean read fOnly8BitSound write fOnly8BitSound;
    function PlayWaveFile( const Filename: String; Volume, Pan: Integer ): PWave;
    function PlayWaveResource( Inst: Integer; RsrcName: PChar; Volume, Pan: Integer ): PWave;
    function PlayWaveMemory( Mem: Pointer; Volume, Pan: Integer ): PWave;
    property Prty: Integer read fPrty write SetPrty;
  end;

var
  MultiWavObj: PMultiWav;
  SoundFreq: Integer = 44;
  SoundBits: Integer = 16;
  NonExactAlreadyRun: Boolean;
  DoNotResetNonExactAlreadyRun: Boolean;

implementation

procedure DDLog( const S: String );
begin
  //todo: implement, please
end;

function CoopLevel( f: DWORD ): String;
begin
  if f = DSSCL_EXCLUSIVE then
    Result := 'DSSCL_EXCLUSIVE'
  else
  if f = DSSCL_NORMAL then
    Result := 'DSSCL_NORMAL'
  else
  if f = DSSCL_PRIORITY then
    Result := 'DSSCL_PRIORITY'
  else
  if f = DSSCL_WRITEPRIMARY then
    Result := 'DSSCL_WRITEPRIMARY'
  else
    Result := '$' + Int2Hex( f, 4 );
end;

{ TMultiWav }

function TMultiWav.CreateSndBuffer(var Buf: IDirectSoundBuffer; nChan,
  nHerz, nBits: Integer; var Format: TWaveFormatEx): Boolean;
var DSBufDesc: TDSBufferDesc;
    WF: TWaveFormatEx;
    R: HResult;
    F: DWORD;
label Try_CreateSoundBuffer;
begin
  Result := FALSE;
  FillChar( WF, Sizeof( WF ), 0 );
  WF.wFormatTag := WAVE_FORMAT_PCM;
  WF.nChannels := nChan;
  WF.nSamplesPerSec := nHerz;
  WF.nBlockAlign := nChan * nBits div 8;
  WF.nAvgBytesPerSec := WF.nBlockAlign * WF.nSamplesPerSec;
  WF.wBitsPerSample := nBits;

  F := DSSCL_PRIORITY;
Try_CreateSoundBuffer:
  FillChar( DSBufDesc, Sizeof( DSBufDesc ), 0 );
  DSBufDesc.dwSize := Sizeof( DSBufDesc );
  DSBufDesc.dwFlags := DSBCAPS_GLOBALFOCUS or
    DSBCAPS_GETCURRENTPOSITION2 or DSBCAPS_STATIC;
  if FPrimaryBuffer then
    DSBufDesc.dwFlags := DSBufDesc.dwFlags or DSBCAPS_PRIMARYBUFFER
                      or DSBCAPS_GETCURRENTPOSITION2 or DSBCAPS_STICKYFOCUS
                      or DSBCAPS_STATIC
  else
  if F = DSSCL_NORMAL then
    DSBufDesc.dwFlags := 0;
  DSBufDesc.dwBufferBytes := DWORD( BufFrames ) * WF.nAvgBytesPerSec div 25;
  BufSize := DSBufDesc.dwBufferBytes;
  if FPrimaryBuffer then
    DSBufDesc.dwBufferBytes := 0;
  DSBufDesc.lpwfxFormat := @ WF;
  R := DS.CreateSoundBuffer( DSBufDesc, Buf, nil );
  if (R <> DS_OK) and FPrimaryBuffer then
  begin
    DDLog( 'Could not create Direct X sound buffer. Error ' + Int2Hex( R, 8 ) + #13#10 +
            DsErrorString( R ) );
    FPrimaryBuffer := FALSE;
    if ExclusiveSound then
      F := DSSCL_EXCLUSIVE;
    R := DS.SetCooperativeLevel( Applet.Handle, F );
    if R <> DS_OK then
      DDLog( 'Could not set cooperative level: ' + CoopLevel( F ) + ', error ' +
        Int2Hex( R, 8 ) + #13#10 + DsErrorString( R ) );
    goto Try_CreateSoundBuffer;
  end
    else
  if (R <> DS_OK) and not FPrimaryBuffer and (F <> DSSCL_NORMAL) then
  begin
    F := DSSCL_NORMAL;
    R := DS.SetCooperativeLevel( Applet.Handle, F );
    if R <> DS_OK then
      DDLog( 'Could not set cooperative level: ' + CoopLevel( F ) + ', error ' +
        Int2Hex( R, 8 ) + #13#10 + DsErrorString( R ) );
    goto Try_CreateSoundBuffer;
  end;
  if R = DS_OK then
  begin
    Buf.GetFormat( @ Format, Sizeof( Format ), nil );
    Result := TRUE;
    IgnorePosition := BufSize;
  end
    else
  begin
    DDLog( 'Could not create Direct X sound buffer. Error ' + Int2Hex( R, 8 ) + #13#10 +
            DsErrorString( R ) );
  end;
end;

destructor TMultiWav.Destroy;
begin
  fDestroying := TRUE;
  Enabled := FALSE;
  if fThread <> nil then
    fThread.WaitFor;
  fThread.Free;
  DeleteCriticalSection( fCrit );
  fWaves.ReleaseObjects;
  Sampling := FALSE;
  SB := nil;
  DS := nil;
  if Buffer <> nil then
  begin
    FreeMem( Buffer );
    Buffer := nil;
  end;
  inherited;
end;

function TMultiWav.FillBuffer_16: Boolean;
var i, n, m: Integer;
    W: PWave;
    AvgBuf: array[ 0..3527 ] of DWord;
    P: PInteger;
    P1: ^SmallInt;
    V: Integer;
begin
  Result := FALSE;
  n := 0;
  for i := 0 to fWaves.Count-1 do
  begin
    W := fWaves.Items[ i ];
    if not W.Eof and not W.Paused then
    begin
      Inc( n );
      if W.FFadeCount > 0 then
      begin
        Dec( W.FFadeCount );
        Inc( W.Volume, W.FFadeDelta );
      end;
      W.ReadSingleBuffer;
    end;
  end;
  if n <> 0 then
  begin

    //       
    FillChar( AvgBuf, Sizeof( AvgBuf ), 0 );
    for i := 0 to fWaves.Count-1 do
    begin
      P := @ AvgBuf[ 0 ];
      W := fWaves.Items[ i ];
      if W.Paused then continue;
      P1 := @ W.fWavBuffer[ 0 ];
      for m := 0 to (High( AvgBuf )+1)div 2 - 1 do
      begin
        V := P1^ * W.Volume div 100;
        Inc( P1 );
        if W.Pan <> 0 then
          P^ := P^ + V * (100-W.Pan) div 100
             + P1^ * W.Volume * Abs(W.Pan) div 10000
        else
          P^ := P^ + V;
        Inc( P );
        V := P1^ * W.Volume div 100;
        if W.Pan <> 0 then
        begin
          Dec( P1 );
          P^ := P^ + V * (100+W.Pan) div 100
             + P1^ * W.Volume * Abs(W.Pan) div 10000;
          Inc( P1 );
        end
          else
          P^ := P^ + V;
        Inc( P );
        Inc( P1 );
      end;
    end;
    P := @ AvgBuf[ 0 ];
    for m := 0 to High( AvgBuf )-1 do
    begin
      //P^ := P^ div n;
      if P^ < -32767 then
        P^ := -32767
      else
      if P^ > 32767 then
        P^ := 32767;
      Inc( P );
    end;

    // ,     1/25    
    if (WFormat.nChannels = 2) and (WFormat.nSamplesPerSec = 44100) and
       (WFormat.wBitsPerSample = 16) then
    begin
      P := @ AvgBuf[ 0 ];
      P1 := Pointer( Buffer );
      for m := 0 to High( AvgBuf ) do
      begin
        PWord( P1 )^ := P^;
        Inc( P );
        Inc( P1 );
      end;
    end;

    Result := TRUE;
  end
    else
  begin
    FillChar( Buffer^, BufLen, 0 );
  end;

  //   
  for i := fWaves.Count-1 downto 0 do
  begin
    W := fWaves.Items[ i ];
    if W.Eof then
    begin
      fWaves.Delete( i );
      W.Free;
    end;
  end;

  if Assigned( OnEndFrame ) then
    OnEndFrame( @ Self );
end;

procedure TMultiWav.GenerateSample;

  procedure DoGenerateSample;
  begin
    if Enabled then
      FillBuffer_16;
  end;

var OldPos: Integer;
begin
  if not FEnabled then Exit;

  if not Sampling then
  begin
    if not AppletTerminated and not fDestroying then
    begin
      IgnorePosition := BufSize;
      CheckPosition := 0;
    end
    else
      Exit;
  end;

  Sampling := TRUE;
  OldPos := CheckPosition;

  DoGenerateSample;

  StartSample;
  CheckPosition := CPosition - (BufLen - 1);
  if CheckPosition > BufSize then
    Dec( CheckPosition, BufSize );
  if CheckPosition < 0 then
    CheckPosition := CheckPosition + BufSize; //BufSize - Integer( BufLen - 1 );
  if CheckPosition < OldPos then
    IgnorePosition := OldPos
  else
    IgnorePosition := BufSize;
end;

procedure TMultiWav.Init;
var R: HResult;
    F: DWORD;
    I: Integer;
begin
  inherited;

  CPosition := $FFFFFFFF;
  R := DirectSoundCreate( nil, DS, nil );
  if R <> 0 then
  begin
    DDLog( 'Could not create Direct Sound Interface, error ' + Int2Str( R ) +
                 ': ' + DsErrorString( R ) );
    Exit;
  end;
  if R = 0 then
  begin
    // created OK, initialize
    F := DSSCL_PRIORITY;
    //F := DSSCL_NORMAL;
    if ExclusiveSound then
      F := DSSCL_EXCLUSIVE;
    {R := DS.SetCooperativeLevel( Applet.GetWindowHandle, DSSCL_WRITEPRIMARY );
    if R = 0 then
      FPrimaryBuffer := TRUE
    else}
      R := DS.SetCooperativeLevel( Applet.GetWindowHandle, F );
    if R <> DS_OK then
    begin
      DDLog( 'Could not set cooperative level: ' + CoopLevel( F ) + ', error ' +
        Int2Hex( R, 8 ) + #13#10 + DsErrorString( R ) );
    end;;
    WFormat.cbSize := Sizeof( WFormat );

    I := 44100;
    CASE SoundFreq OF
    44: I := 44100;
    22: I := 22050;
    11: I := 11025;
    8:  I := 8000;
    END;

    if not CreateSndBuffer( SB, 2, I, SoundBits, WFormat ) then
    begin

      if not Only8BitSound then
      if not CreateSndBuffer( SB, 2, 44100, 16, WFormat ) then
      if not CreateSndBuffer( SB, 2, 22050, 16, WFormat ) then
      if not CreateSndBuffer( SB, 2, 11025, 16, WFormat ) then;

      if not Assigned( SB ) then
      if not CreateSndBuffer( SB, 2, 8000, 8, WFormat ) then
      if not CreateSndBuffer( SB, 2, 11025, 8, WFormat ) then
      if not CreateSndBuffer( SB, 2, 22050, 8, WFormat ) then
      if not CreateSndBuffer( SB, 2, 44100, 8, WFormat ) then
      ;

    end;

    if Assigned( SB ) then
    begin
      BufLen := WFormat.nAvgBytesPerSec div 25;
      //BufSize := BufLen * 4 + 8;
      Buffer := AllocMem( BufLen + 16 );
      FEnabled := TRUE;
      IgnorePosition := BufSize;
    end;

  end;
  fWaves := NewList;
  fPrty := NORMAL_PRIORITY_CLASS;
  InitializeCriticalSection( fCrit );
end;

procedure TMultiWav.LogFile1( Value: DWORD );
begin
  //
end;

function TMultiWav.OK: Boolean;
begin
  TRY
    Result := Assigned( DS ) and Assigned( SB );
  EXCEPT
    Result := FALSE;
  END;
end;

function TMultiWav.PlayWaveFile(const Filename: String; Volume,Pan: Integer): PWave;
var W: PWave;
begin
  Result := nil;
  EnterCriticalSection( fCrit );
  TRY
    new( W, Create );
    W.fMultiWav := @ Self;
    if not W.LoadFile( Filename ) then
    begin
      W.Free;
      Exit;
    end;
    fWaves.Add( W );
    W.Pan := Pan;
    W.Volume := Volume;
    Stopped := FALSE;
    if (fThread = nil) or not Sampling then
      StartSampling;
    Result := W;
  FINALLY
    LeaveCriticalSection( fCrit );
  END;
end;

function TMultiWav.PlayWaveMemory(Mem: Pointer; Volume,Pan: Integer): PWave;
var W: PWave;
begin
  Result := nil;
  EnterCriticalSection( fCrit );
  TRY
    new( W, Create );
    W.fMultiWav := @ Self;
    W.fMem := NewExMemoryStream( Mem, PDWORD( DWORD( Mem ) + 4 )^ );
    if not W.LoadMemory then
    begin
      W.Free;
      Exit;
    end;
    fWaves.Add( W );
    W.Pan := Pan;
    W.Volume := Volume;
    Stopped := FALSE;
    if (fThread = nil) or not Sampling then
      StartSampling;
    Result := W;
  FINALLY
    LeaveCriticalSection( fCrit );
  END;
end;

function TMultiWav.PlayWaveResource(Inst: Integer; RsrcName: PChar;
  Volume,Pan: Integer): PWave;
var W: PWave;
begin
  Result := nil;
  EnterCriticalSection( fCrit );
  TRY
    new( W, Create );
    W.fMultiWav := @ Self;
    fWaves.Add( W );
    if not W.LoadRsrc( Inst, RsrcName ) then
    begin
      W.Free;
      Exit;
    end;
    W.Pan := Pan;
    W.Volume := Volume;
    Stopped := FALSE;
    if (fThread = nil) or not Sampling then
      StartSampling;
    Result := W;
  FINALLY
    LeaveCriticalSection( fCrit );
  END;
end;

procedure TMultiWav.SetEnabled(const Value: Boolean);
begin
  FEnabled := Value;
  if not Value then Stop;
end;

procedure TMultiWav.SetPrty(const Value: Integer);
begin
  fPrty := Value;
  if fThread <> nil then
    fThread.PriorityClass := fPrty;
end;

procedure TMultiWav.StartSample;
var L, L2: DWORD;
    AudioPos1, AudioPos2: Pointer;
    AudioLen1, AudioLen2: DWORD;
    Status: DWORD;
    Buf: PByte;
    R: Integer;

    P1: Integer;
label Try_Lock;
begin
  if not FEnabled then Exit;
  if Sampling then
  begin
      Buf := Buffer;
      if CPosition = $FFFFFFFF then
        //SB.GetCurrentPosition( nil, @ CPosition );
      begin
        CPosition := 0;
      end;
      //----------------------------------------------------------------------//
      Try_Lock:
      R := SB.Lock( CPosition, BufLen, AudioPos1, AudioLen1, AudioPos2, AudioLen2,
               0 );

      P1 := CPosition;

      if R = DS_OK then
      begin
        L := Min( AudioLen1, BufLen );
        if (AudioPos1 <> nil) and (L > 0) then
          Move( Buf^, AudioPos1^, L );
        L2 := 0;
        if (AudioPos2 <> nil) and (AudioLen2 > 0) and
           (L < BufLen) then
        begin
          L2 := BufLen - L;
          Move( PByte( DWORD( Buffer ) + L )^, AudioPos2^, Min( L2, AudioLen2 ) );
        end;
        SB.Unlock( AudioPos1, L, AudioPos2, L2 );
        //Inc( Buf, BufLen );
        Inc( CPosition, L+L2 );

      end
        else
      begin
        if R = DSERR_BUFFERLOST then
        begin
          R := SB.Restore;
          if R = DS_OK then
            goto Try_Lock;
        end;
      end;
      if CPosition >= DWORD( BufFrames ) * BufLen  then
        Dec( CPosition, DWORD( BufFrames )* BufLen  );

      if SB.GetStatus( Status ) = 0 then
      begin
        if not LongBool( Status and DSBSTATUS_PLAYING ) then
        begin
          SB.SetCurrentPosition( P1 );
          SB.Play( 0, 0, DSBPLAY_LOOPING );
        end;
      end;

  end
    else
  begin
    if not OK then Exit;
  end;
end;

procedure TMultiWav.StartSampling;
begin
  if fThread = nil then
  begin
    fThread := NewThread;
    fThread.PriorityClass := fPrty;
    fThread.OnExecute := ThreadExecute;
    fThread.Resume;
  end;
end;

procedure TMultiWav.Stop;
begin
  EnterCriticalSection( fCrit );
  TRY
    Stopped := TRUE;
    if Assigned( SB ) then
      SB.Stop;
    Sampling := FALSE;
    CPosition := $FFFFFFFF;
    CheckPosition := 0;
    IgnorePosition := BufSize;
    if Buffer <> nil then
      FillChar( Buffer^, BufLen, 0 );
    fWaves.ReleaseObjects;
    fWaves := NewList;
  FINALLY
    LeaveCriticalSection( fCrit );
  END;
end;

function TMultiWav.ThreadExecute(Sender: PThread): Integer;
var Read: DWORD;
begin
  while not AppletTerminated and not fDestroying do
  begin
    if not Stopped then
    begin
      SB.GetCurrentPosition( nil, @ Read );
      if (Integer( Read ) < IgnorePosition) and (Integer( Read ) >= CheckPosition)
         or not Sampling then
      begin
        //Sampling := TRUE;
        EnterCriticalSection( fCrit );
        TRY
          GenerateSample;
        FINALLY
          LeaveCriticalSection( fCrit );
        END;
      end;
    end;
    Sleep( 12 );
  end;
  Sampling := FALSE;
  SB.Stop;
  Result := 0;
end;

{ TWave }

function TWave.CheckSignature( const Signature: TWaveFileHeader ): Boolean;
begin
  Result := FALSE;
  if (Signature.lRIFF <> 'RIFF') or (Signature.lWave <> 'WAVE') or
     (Signature.lFmtLen < 16) then
    Exit;
  Result := TRUE;
end;

function TWave.CheckWaveFormatPCM: Boolean;
begin
  Result := (fWaveFormat.wFormatTag=WAVE_FORMAT_PCM) and
            IntIn( fWaveFormat.nChannels, [ 1, 2 ] ) and
            //IntIn( fWaveFormat.nSamplesPerSec, [ 44100, 22050, 11025 {, 8000} ] ) and
            IntIn( fWaveFormat.wBitsPerSample, [ 8, 16 ] );
end;

procedure TWave.Close;
begin
  if fFile <> INVALID_HANDLE_VALUE then
    CloseHandle( fFile );
  fFile := 0;
  Free_And_Nil( fMem );
  fSize := 0;
end;

destructor TWave.Destroy;
begin
  Close;
  inherited;
end;

function TWave.Eof: Boolean;
begin
  Result := TRUE;
  if not OK then Exit;
  Result := fSize <= 0;
end;

function TWave.LoadFile(const Filename: String): Boolean;
var Signature: TWaveFileHeader;
    Chunk: TWaveChunk;
begin
  Result := FALSE;
  fFile := FileCreate( Filename, ofOpenRead or ofOpenExisting or ofShareDenyWrite );
  if fFile <> INVALID_HANDLE_VALUE then
  begin
    fSize := GetFileSize( fFile, nil ) - 12 - sizeof( TWaveFormatEx );
    if fSize <= 0 then
    begin
      Close;
      Exit;
    end;
    FileRead( fFile, Signature, Sizeof( Signature ) );
    if not CheckSignature( Signature ) then
    begin
      Close;
      Exit;
    end;
    if FileRead( fFile, fWaveFormat, Sizeof( fWaveFormat ) ) <> Sizeof( fWaveFormat ) then
    begin
      Close;
      Exit;
    end;
    if not CheckWaveFormatPCM then
    begin
      Close;
      Exit;
    end;
    FileSeek( fFile, Signature.lFmtLen-16-2, spCurrent );
    Chunk.lType := 0;
    REPEAT
      if FileRead( fFile, Chunk, Sizeof( Chunk ) ) <> Sizeof( Chunk ) then break;
      if Chunk.lType <> $61746164 then
        FileSeek( fFile, Chunk.lLen, spCurrent );
    UNTIL Chunk.lType = $61746164;
    if Chunk.lType = $61746164 then
      fSize := Chunk.lLen
    else
    begin
      Close;
      Exit;
    end;
    Result := TRUE;
    fStartPosition := FileSeek( fFile, 0, spCurrent );
    fStartSize := fSize;
  end;
end;

function TWave.LoadMemory: Boolean;
var Signature: TWaveFileHeader;
    Chunk: TWaveChunk;
begin
  Result := FALSE;
  if fMem.Size <= 12 then
    Exit;
  fMem.Read( Signature, Sizeof( Signature ) );
  if not CheckSignature( Signature ) then
    Exit;
  if fMem.Read( fWaveFormat, Sizeof( fWaveFormat ) ) <> Sizeof( fWaveFormat ) then
    Exit;
  if not CheckWaveFormatPCM then
    Exit;
  fMem.Position := fMem.Position + DWORD( Signature.lFmtLen-16 )-2;
  Chunk.lType := 0;
  REPEAT
    if fMem.Read( Chunk, Sizeof( Chunk ) ) <> Sizeof( Chunk ) then break;
      if Chunk.lType <> $61746164 then
        fMem.Position := fMem.Position + DWORD( Chunk.lLen );
  UNTIL Chunk.lType = $61746164;
  if Chunk.lType = $61746164 then
  begin
    fSize := Chunk.lLen;
    fStartPosition := fMem.Position;
    fStartSize := fSize;
    Result := TRUE;
  end
  else
    Exit;
end;

function TWave.LoadRsrc(Inst: Integer; RsrcName: PChar): Boolean;
begin
  Result := FALSE;
  fMem.Free;
  fMem := NewMemoryStream;
  Resource2Stream( fMem, Inst, RsrcName, RT_RCDATA );
  if not LoadMemory then
  begin
    Close;
    Exit;
  end;
  Result := TRUE;
end;

function TWave.OK: Boolean;
begin
  Result := (fFile <> 0) and (fFile <> INVALID_HANDLE_VALUE) or
            (fMem <> nil);
end;

procedure TWave.PlayRange(FromSample, Samples: DWORD);
var nBytesPerSample: DWORD;
begin
  Restart;
  nBytesPerSample := fWaveFormat.nChannels * (fWaveFormat.wBitsPerSample div 8);
  if fMem <> nil then
    fMem.Position := fStartPosition + FromSample * nBytesPerSample
  else
  if (fFile <> 0) and (fFile <> INVALID_HANDLE_VALUE) then
    FileSeek( fFile, fStartPosition + FromSample * nBytesPerSample, spBegin )
  else
    Exit;
  fSize := Samples * nBytesPerSample;
  if  fSize > Integer( fStartSize - (fStartPosition + FromSample * nBytesPerSample) ) then
    fSize := fStartSize - (fStartPosition + FromSample * nBytesPerSample);
  if fSize > 0 then
    SetPaused( FALSE );
end;

procedure TWave.ReadSingleBuffer;
var TmpBuf: array[ 0..40959 ] of Byte;
    TmpSize: Integer;
    PDst: PSmallInt;
    PSrc: PByte;
    m, step, speedstep: Integer;
    CurSrcPos, CurSrcPos1, NewSrcPos, DeltaSrcPos: Integer;
    SpeedAlign, SpeedMul: DWORD;
begin
  if not Eof then
  begin

    CASE fWaveFormat.nSamplesPerSec OF
    44100: TmpSize := 44100 div 25 * 4;
    22050: TmpSize := 22050 div 25 * 4;
    11025: TmpSize := 11025 div 25 * 4;
    8000: TmpSize := 8000 div 25 * 4;
    else  TmpSize := fWaveFormat.nSamplesPerSec div 25 * 4;
    END;
    if fWaveFormat.nChannels < 2 then
      TmpSize := TmpSize div 2;
    if fWaveFormat.wBitsPerSample < 16 then
      TmpSize := TmpSize div 2;

    if (TmpSize = 44100 div 25 * 4) and (Speed = wsNormal) then
    begin

      FillChar( fWavBuffer, TmpSize + 16, 0 );
      if (fFile <> 0) and (fFile <> INVALID_HANDLE_VALUE) then
        BufferActualSize := FileRead( fFile, fWavBuffer, Min( fSize, TmpSize ) )
      else
        BufferActualSize := fMem.Read( fWavBuffer, Min( fSize, TmpSize ) );

      if (fSize < TmpSize) and Loop then
      begin
        Restart;
        if (fFile <> 0) and (fFile <> INVALID_HANDLE_VALUE) then
          Inc( BufferActualSize, FileRead( fFile,
            Pointer( Integer( @ fWavBuffer[ 0 ] ) + fSize )^, TmpSize - fSize ) )
        else
          BufferActualSize := fMem.Read( Pointer( Integer( @ fWavBuffer[ 0 ] ) + fSize )^, TmpSize - fSize );
      end
        else
      if (fSize < TmpSize) and (fNextWave <> nil) then
      begin
        fNextWave.Restart;
        if (fNextWave.fFile <> 0) and (fNextWave.fFile <> INVALID_HANDLE_VALUE) then
          Inc( BufferActualSize, FileRead( fNextWave.fFile,
            Pointer( Integer( @ fWavBuffer[ 0 ] ) + fSize )^, TmpSize - fSize ) )
        else
          BufferActualSize := fNextWave.fMem.Read(
            Pointer( Integer( @ fWavBuffer[ 0 ] ) + fSize )^, TmpSize - fSize );
        Dec( fNextWave.fSize, TmpSize - fSize );
        //fNextWave.Paused := FALSE;
      end
        else
      if (fSize < TmpSize) then
      begin
        asm
          nop
        end;
      end;

    end
      else
    begin
      CASE Speed OF
      wsNormal: ;
      wsFast2:  TmpSize := TmpSize * 2;
      wsFast4:  TmpSize := TmpSize * 4;
      wsSlow2:  TmpSize := TmpSize div 2;
      wsSlow4:  TmpSize := TmpSize div 4;
      END;
      SpeedAlign := fWaveFormat.nChannels * (fWaveFormat.wBitsPerSample div 8);
      TmpSize := TmpSize and not (SpeedAlign - 1);
      SpeedMul := 1;
      CASE Speed OF
      wsNormal: begin SpeedAlign := 4; SpeedMul := 1; end;
      wsSlow2:  begin SpeedAlign := 8; SpeedMul := 1; end;
      wsSlow4:  begin SpeedAlign := 16; SpeedMul := 1; end;
      wsFast2:  begin SpeedAlign := 2; SpeedMul := 2; end;
      wsFast4:  begin SpeedAlign := 1; SpeedMul := 4; end;
      END;

      FillChar( TmpBuf, TmpSize, 0 );
      if (fFile <> 0) and (fFile <> INVALID_HANDLE_VALUE) then
        BufferActualSize := FileRead( fFile, TmpBuf, Min( fSize, TmpSize ) )
      else
        BufferActualSize := fMem.Read( TmpBuf, Min( fSize, TmpSize ) );

      if (fSize < TmpSize) and Loop then
      begin
        Restart;
        if (fFile <> 0) and (fFile <> INVALID_HANDLE_VALUE) then
          Inc( BufferActualSize, FileRead( fFile, fWavBuffer, TmpSize - fSize ) )
        else
          BufferActualSize := fMem.Read( fWavBuffer, TmpSize - fSize );
      end;

      PSrc := @ TmpBuf[ 0 ];
      PDst := @ fWavBuffer[ 0 ];
      if (fWaveFormat.nSamplesPerSec <= 44100) and
         (44100 mod fWaveFormat.nSamplesPerSec = 0) then
      begin
        step := 0;
        speedstep := 0;
        for m := 0 to 44100 div 25 - 1 do
        begin
          if speedstep <= 0 then
            speedstep := SpeedAlign div 4;
          if step = 0 then
            step := 44100 div fWaveFormat.nSamplesPerSec;
          if fWaveFormat.wBitsPerSample = 16 then
          begin
            PDst^ := PWord( PSrc )^;
            Inc( PDst );
            if fWaveFormat.nChannels = 2 then
              Inc( PSrc, 2 );
            PDst^ := PWord( PSrc )^;
            if fWaveFormat.nChannels = 2 then
              Dec( PSrc, 2 );
            Dec( speedstep );
            if speedstep <= 0 then
            begin
              Dec( Step );
              if Step = 0 then
                Inc( PSrc, fWaveFormat.nChannels * fWaveFormat.wBitsPerSample div 8
                   * SpeedMul );
            end;
          end
            else
          if fWaveFormat.wBitsPerSample = 8 then
          begin
            PDst^ := (PSrc^ - 128) shl 7;
            Inc( PDst );
            if fWaveFormat.nChannels = 2 then
              Inc( PSrc );
            PDst^ := (PSrc^ - 128) shl 7;
            if fWaveFormat.nChannels = 2 then
              Dec( PSrc );
            Dec( speedstep );
            if speedstep <= 0 then
            begin
              Dec( Step );
              if Step = 0 then
                Inc( PSrc, fWaveFormat.nChannels * fWaveFormat.wBitsPerSample div 8
                  * SpeedMul );
            end;
          end;
          Inc( PDst );
        end;
      end
        else
      begin //      44100 stereo 16 bit
        FillChar( fWavBuffer, Sizeof( fWavBuffer ), 0 );
        CurSrcPos := 0;
        DeltaSrcPos := fWaveFormat.nSamplesPerSec * 16384 div 44100 * SpeedAlign div 4;
        for m := 0 to 44100 div 25 - 1 do
        begin
          NewSrcPos := CurSrcPos + DeltaSrcPos;
          CurSrcPos1 := CurSrcPos;
          if fWaveFormat.wBitsPerSample = 16 then
          begin
            PSrc := @ TmpBuf[ 2 * fWaveFormat.nChannels * (CurSrcPos div 16384) ];
            while CurSrcPos1 < NewSrcPos do
            begin
              if (CurSrcPos1 + DeltaSrcPos) div 16384  > CurSrcPos1 div 16384 then
              begin
                PDst^ := PDst^ + PSmallInt( PSrc )^ * (16384 - CurSrcPos1 and 16383) div DeltaSrcPos;
                Inc( PSrc, 2 * fWaveFormat.nChannels );
                Inc( CurSrcPos1, 16384 - CurSrcPos1 and 16383 );
              end
              else
              if DeltaSrcPos < 16384 then
              begin
                PDst^ := PDst^ + PSmallInt( PSrc )^  * (NewSrcPos - CurSrcPos1) div DeltaSrcPos;
                Inc( CurSrcPos1, 16384 );
              end
                else
              begin
                PDst^ := PDst^ + PSmallInt( PSrc )^ * 16384 div DeltaSrcPos;
                Inc( CurSrcPos1, 16384 );
                Inc( PSrc, 2 * fWaveFormat.nChannels );
              end;
            end;
          end
            else
          if fWaveFormat.wBitsPerSample = 8 then
          begin
            PSrc := @ TmpBuf[ fWaveFormat.nChannels * (CurSrcPos div 16384) ];
            while CurSrcPos1 < NewSrcPos do
            begin
              if (CurSrcPos1 + DeltaSrcPos) div 16384  > CurSrcPos1 div 16384 then
              begin
                PDst^ := PDst^ + (PSrc^-128) * (16384 - CurSrcPos1 and 16383) div DeltaSrcPos * 128;
                Inc( PSrc, 2 * fWaveFormat.nChannels );
                Inc( CurSrcPos1, 16384 - CurSrcPos1 and 16383 );
              end
              else
              if DeltaSrcPos < 16384 then
              begin
                PDst^ := PDst^ + (PSrc^-128) * (NewSrcPos - CurSrcPos1) div DeltaSrcPos * 128;
                Inc( CurSrcPos1, 16384 );
              end
                else
              begin
                PDst^ := PDst^ + (PSrc^-128) * 16384 div DeltaSrcPos * 128;
                Inc( CurSrcPos1, 16384 );
                Inc( PSrc, fWaveFormat.nChannels );
              end;
            end;
          end;
          Inc( PDst );
          CurSrcPos1 := CurSrcPos;
          if fWaveFormat.wBitsPerSample = 16 then
          begin
            if fWaveFormat.nChannels = 2 then
              PSrc := @ TmpBuf[ 2 * fWaveFormat.nChannels * (CurSrcPos div 16384) + 2 ]
            else
              PSrc := @ TmpBuf[ 2 * fWaveFormat.nChannels * (CurSrcPos div 16384) ];
            while CurSrcPos1 < NewSrcPos do
            begin
              if (CurSrcPos1 + DeltaSrcPos) div 16384  > CurSrcPos1 div 16384 then
              begin
                PDst^ := PDst^ + PSmallInt( PSrc )^ * (16384 - CurSrcPos1 and 16383) div DeltaSrcPos;
                Inc( PSrc, 2 * fWaveFormat.nChannels );
                Inc( CurSrcPos1, 16384 - CurSrcPos1 and 16383 );
              end
              else
              if DeltaSrcPos < 16384 then
              begin
                PDst^ := PDst^ + PSmallInt( PSrc )^  * (NewSrcPos - CurSrcPos1) div DeltaSrcPos;
                Inc( CurSrcPos1, 16384 );
              end
                else
              begin
                PDst^ := PDst^ + PSmallInt( PSrc )^ * 16384 div DeltaSrcPos;
                Inc( CurSrcPos1, 16384 );
                Inc( PSrc, 2 * fWaveFormat.nChannels );
              end;
            end;
          end
            else
          if fWaveFormat.wBitsPerSample = 8 then
          begin
            if fWaveFormat.nChannels = 2 then
              PSrc := @ TmpBuf[ fWaveFormat.nChannels * (CurSrcPos div 16384) + 1 ]
            else
              PSrc := @ TmpBuf[ fWaveFormat.nChannels * (CurSrcPos div 16384) ];
            while CurSrcPos1 < NewSrcPos do
            begin
              if (CurSrcPos1 + DeltaSrcPos) div 16384  > CurSrcPos1 div 16384 then
              begin
                PDst^ := PDst^ + (PSrc^-128) * (16384 - CurSrcPos1 and 16383) div DeltaSrcPos * 128;
                Inc( PSrc, 2 * fWaveFormat.nChannels );
                Inc( CurSrcPos1, 16384 - CurSrcPos1 and 16383 );
              end
              else
              if DeltaSrcPos < 16384 then
              begin
                PDst^ := PDst^ + (PSrc^-128) * (NewSrcPos - CurSrcPos1) div DeltaSrcPos * 128;
                Inc( CurSrcPos1, 16384 );
              end
                else
              begin
                PDst^ := PDst^ + (PSrc^-128) * 16384 div DeltaSrcPos * 128;
                Inc( CurSrcPos1, 16384 );
                Inc( PSrc, fWaveFormat.nChannels );
              end;
            end;
          end;
          Inc( PDst );
          CurSrcPos := NewSrcPos;
        end;
      end;
    end;
    Dec( fSize, TmpSize );
  end;
  if Eof and Assigned( OnEof ) then
    OnEof( @ Self );
  if Eof and Loop then
    Restart;
end;

procedure TWave.Restart;
begin
  if fMem <> nil then
    fMem.Position := fStartPosition
  else
  if (fFile <> 0) and (fFile <> INVALID_HANDLE_VALUE) then
    FileSeek( fFile, fStartPosition, spBegin )
  else
    Exit;
  fSize := fStartSize;
  SetPaused( FALSE );
end;

function TWave.SetFade(FadeDelta, FadeCount: Integer): PWave;
begin
  FFadeDelta := FadeDelta;
  FFadeCount := FadeCount;
  Result := @ Self;
end;

procedure TWave.SetPaused(const Value: Boolean);
begin
  fPaused := Value;
  if not Value then
  begin
    if fMultiWav.fWaves.IndexOf( @ Self ) < 0 then
      fMultiWav.fWaves.Add( @ Self );
  end;
end;

procedure TWave.Stop;
begin
  if @ Self = nil then Exit;
  fSize := 0;
end;

function TWave.WaveFormat: TWaveFormatEx;
begin
  Result := fWaveFormat;
end;

end.
